import logging
from logging import config

log_config = {
    "version": 1,
    "root": {"handlers": ["console"], "level": "DEBUG"},
    "handlers": {
        "console": {
            "formatter": "std_out",
            "class": "logging.StreamHandler",
            "level": "DEBUG",
        }
    },
    "formatters": {
        "std_out": {
            "format": "%(asctime)s | %(levelname)s | %(module)s | %(funcName)s | %(message)s",
            "datefmt": "%d-%m-%Y %H:%M:%S",
        }
    },
}

config.dictConfig(log_config)

logger = logging.getLogger("portfolio")



import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


def plot_histogram(
    returns: pd.Series,
    percent_returns: pd.Series,
    plot_title: str,
    save: bool = False,
) -> None:
    axs = plt.figure(constrained_layout=True).subplots(1, 2)
    plots = [
        (axs[0], "Returns", returns, False),
        (axs[1], "Percent Returns", percent_returns, True),
    ]
    for ax, title, data, density in plots:
        ax.set(title=title)
        ax.hist(data, bins=75, density=density)
    if save:
        plt.savefig(plot_title)
    plt.show()


def plot_heatmap(matrix: pd.DataFrame, plot_title: str, save: bool = False) -> None:
    sns.heatmap(
        matrix,
        annot=True,
        cmap="YlGnBu",
        linewidth=0.3,
        annot_kws={"size": 9},
    )
    plt.xticks(rotation=90)
    plt.yticks(rotation=0)
    if save:
        plt.savefig(plot_title)
    plt.show()



import typing

import numpy as np
import pandas as pd

# import statsmodels.formula.api as smf
from scipy import stats


def calculate_mean_daily_returns(clean_returns: pd.Series) -> float:
    return np.mean(clean_returns)


def calculate_mean_annual_return(clean_returns: pd.Series) -> float:
    mean_returns = calculate_mean_daily_returns(clean_returns)
    return ((1 + mean_returns) ** 252) - 1


def calculate_std_daily(clean_returns: pd.Series) -> float:
    return np.std(clean_returns)


def calculate_variance_daily(clean_returns: pd.Series) -> float:
    return calculate_std_daily(clean_returns) ** 2


def calculate_skewness(clean_returns: pd.Series) -> float:
    return stats.skew(clean_returns)


def calculate_kurtosis(clean_returns: pd.Series) -> float:
    return stats.kurtosis(clean_returns)


def calculate_shapiro(clean_returns: pd.Series) -> typing.Tuple[float, float]:
    """
    Calculate shapiro and p-value
    """
    shapiro = stats.shapiro(clean_returns)
    return shapiro[0], shapiro[1]


def calculate_percentage_returns(clean_returns: pd.Series) -> float:
    percent = clean_returns * 100
    return percent.dropna()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['invest_tools']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.7.0,<4.0.0',
 'numpy>=1.24.2,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'scipy>=1.10.1,<2.0.0',
 'seaborn>=0.12.2,<0.13.0',
 'statsmodels>=0.13.5,<0.14.0']

setup_kwargs = {
    'name': 'invest-tools',
    'version': '0.2.4',
    'description': 'Tools to manage portfolio risk analysis',
    'long_description': '# invest-tools\n\n[![PyPI version](https://badge.fury.io/py/invest-tools.svg)](https://badge.fury.io/py/invest-tools)\n[![codecov](https://codecov.io/gh/leo-jp-edwards/invest-tools/graph/badge.svg?token=C1W8MZFS80)](https://codecov.io/gh/leo-jp-edwards/invest-tools)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n**Tools to manage portfolio risk analysis**\n\n## Installation\n\nAs a python package this should be installable through:\n\n> `pip install invest-tools`\n\nOr:\n\n> `poetry add invest-tools`\n\n### Dependencies\n\nThe dependencies of this project can be seen in the `pyproject.toml` file. However for completeness there is a dependcy on `pandas`, `statsmodels` and `matplotlib` as the basics.\n\n## Data Inputs\n\nThere are three data inputs which should be present for the package to work as expected. \n\nThe path strings to the csvs can be passed in. \n\n1. Portfolio price data as a CSV\n\n| TIDM | Date | Open | High | Low | Close | Volume | Adjustment |\n|------|------|------|------|-----|-------|--------|------------|\n| EG | 01/01/2023 | 1 | 1 | 1 | 1 | 1 | 1 |\n| EG2 | 01/01/2023 | 1 | 1 | 1 | 1 | 1 | 1 |\n\n2. Currency data as a CSV\n\n| Date | Open | High | Low | Close | Adj Close | Volume |\n|------|------|------|-----|-------|-----------|--------|\n| 01/01/2023 | 1 | 1 | 1 | 1 | 1 | 1 |\n\n3. Fama French Data as a CSV\n\nThis can be found [here](https://mba.tuck.dartmouth.edu/pages/faculty/ken.french/data_library.html)\n\n## Example\n\nBuild a portfolio of two securities called `EG` and `EG2` with the weighting split 50:50 between the two. One is denominated in GBP and one in USD.\n\nThis will output the mean returns of such a portfolio.\n\n```python\nimport numpy as np\nfrom invest_tools import portfolio\n\nportfolio_definition = {\n    "EG": {\n        "weight": 0.5,\n        "currency": "gbp"\n    },\n    "EG2": {\n        "weight": 0.5,\n        "currency": "usd"\n    }\n}\n\nport = portfolio.Portfolio(portfolio_definition, portfolio.Currency.GBP)\nport.get_usd_converter("path/to/csv")\nport.get_prices("path/to/csv")\nport.build()\nport.analyse()\nprint(port.analysis)\nport.plot_correlation_heatmap()\nport.plot_returns_data()\n```\n\n## Roadmap\n\n- [x] Add an input validator\n- [x] Add logging\n- [ ] Add deeper analysis methods\n    - [ ] Coppock Curve\n    - [ ] Fama French\n    - [ ] Excess Returns\n    - [ ] Maximum Drawdown\n    - [ ] Calculate Moments\n- [ ] Add further testing\n- [ ] Make the package more generic\n- [ ] Investigate using [Polars](https://www.pola.rs/)\n\n## License\n\n[MIT](LICENSE)\n\n## Contact\n\nJust open an issue I guess?',
    'author': 'leo',
    'author_email': 'leojpedwards@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)



import math
import typing

import numpy as np
import pandas as pd

from invest_tools.currency import Currency, InvalidCurrencyException


class InvalidDataFrame(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidPortfolioDefinition(Exception):
    def __init__(self, message):
        super().__init__(message)


def validate_columns(df: pd.DataFrame, columns: typing.List[str]) -> bool:
    if set(df.columns) != set(columns):
        diff = []
        for col in set(df.columns):
            if col not in set(columns):
                diff.append(col)
        raise InvalidDataFrame(f"Invalid DataFrame due to column error: {diff}")
    else:
        return True


def validate_datatypes(df: pd.DataFrame, column_types: typing.Dict[str, str]) -> bool:
    try:
        df = df.astype(column_types)
        return True
    except ValueError as e:
        raise InvalidDataFrame(f"Invalid DataFrame due to datatype error: {e}")


def validate_portfolio_definition(
    definition: typing.Dict[str, typing.Dict[str, str]]
) -> bool:
    weights = []
    currencies = set(currency.value for currency in Currency)
    for code, opts in definition.items():
        if "weight" not in opts.keys():
            raise InvalidPortfolioDefinition(f"weight definition missing from {code}")
        if "currency" not in opts.keys():
            raise InvalidPortfolioDefinition(f"currency definition missing from {code}")
        if opts["currency"] not in currencies:
            raise InvalidCurrencyException(f"{opts['currency']} not permitted")
        weights.append(opts["weight"])
    total_weight = np.sum(weights)
    # floating point precsion + rounding means that sometimes this is not 1 exactly
    if not math.isclose(total_weight, 1):
        raise InvalidPortfolioDefinition(f"total weights of {total_weight} should be 1")
    return True



import typing

import pandas as pd

from invest_tools import analysis, currency, plot, validation
from invest_tools.log import logger

PRICES_DATATYPES = {
    "TIDM": "string",
    "Date": "string",
    "Open": float,
    "High": float,
    "Low": float,
    "Close": float,
    "Volume": float,
    "Adjustment": float,
}

CURRENCY_DATATYPES = {
    "Date": "string",
    "Open": float,
    "High": float,
    "Low": float,
    "Close": float,
    "Adj Close": float,
    "Volume": float,
}


class Portfolio:
    """
    An investment portfolio is a collection of securities with weights and
    currencies. Using historical prices these portfolios can be analysed for
    historical performance and benchmarked by different values.

    Analysis can be done using a variety of techniques.

    The main function of building the portfolio is to standardise the portfolio
    to a single currency using a conversion rate (the Close price of the
    relevant currency on the day of the returns).
    """

    def __init__(
        self,
        portfolio_definition: typing.Dict[str, typing.Dict[str, str]],
        currency: currency.Currency,
    ):
        """
        The portfolio definition must be a python dictionary with the form of:

        ```
            {
                "code": {
                    "weight": Float,
                    "currency": "gbp" | "usd"
                },
                ...
            }
        ```

        The sum of the weights must be 1.

        The currency is defined in the `Currency` enum.

        Other values are simply empty initialised for future use.
        """
        validation.validate_portfolio_definition(portfolio_definition)
        logger.info("validated portfolio definition")
        self.portfolio_definition = portfolio_definition
        self.backtest = pd.DataFrame()
        self.prices = pd.DataFrame()
        self.gbpusd = pd.DataFrame()
        self.usdgbp = pd.DataFrame()
        self.currency = currency
        self.clean_returns = pd.Series(dtype=float)
        self.percentage_returns = pd.Series(dtype=float)
        self.analysis = {}

    def ping(self):
        logger.info("PING")
        return "pong"

    def build(self) -> pd.DataFrame:
        """
        Use the portfolio definition to build the portfolio

        :returns: Pandas dataframe of the portfolio returns
        :rtype: pd.DataFrame
        """
        dfs = []
        weights = []
        for code, opts in self.portfolio_definition.items():
            weights.append(opts["weight"])
            if opts["currency"] != self.currency:
                if opts["currency"] != currency.Currency.GBP:
                    ret = self.calculate_returns(
                        self.prices, code, convert=True, cur=self.gbpusd
                    )
                elif opts["currency"] != currency.Currency.USD:
                    ret = self.calculate_returns(
                        self.prices, code, convert=True, cur=self.usdgbp
                    )
            else:
                ret = self.calculate_returns(
                    self.prices, code, convert=False, cur=self.gbpusd
                )
            ret = ret.rename({"Returns": code}, axis=1)
            dfs.append(ret[code].to_frame())
        port = dfs[0].join(dfs[1:])
        port_ret = port.mul(weights, axis=1).sum(axis=1)
        port["portfolio_returns"] = port_ret
        self.backtest = port
        self.clean_returns = port_ret.dropna()
        logger.info("Portfolio built")
        return port

    def get_prices(self, prices_csv: str) -> pd.DataFrame:
        """
        Take in a string pointing to a csv file containing the prices

        The CSV should be in the following format:

        | TIDM | Date | Open | High | Low | Close | Volume | Adjustment |
        |------|------|------|------|-----|-------|--------|------------|

        The Date column should be in the format of "%d/%m/%Y".

        :returns: Pandas dataframe of the portfolio prices
        :rtype: pd.DataFrame
        """
        logger.info(f"Loading data from {prices_csv}")
        df = pd.read_csv(prices_csv)
        valid = validation.validate_columns(df, PRICES_DATATYPES.keys())
        valid = validation.validate_datatypes(df, PRICES_DATATYPES)
        logger.debug(f"passed validation: {valid}")
        df["Date"] = pd.to_datetime(df["Date"], format="%d/%m/%Y")
        self.prices = df
        return df

    # TODO make this generic for currency
    def get_usd_converter(self, conversion_csv: str) -> pd.DataFrame:
        """
        Get a dataframe of USD to GBP to convert the prices between currencies.
        All portfolio prices and returns should be in GBP.

        The CSV should be in the following format:

        | Date | Open | High | Low | Close | Adj Close | Volume |
        |------|------|------|-----|-------|-----------|--------|

        :returns: Pandas dataframe of currency prices.
        :rtype: pd.DataFrame
        """
        logger.info(f"Loading data from {conversion_csv}")
        cur = pd.read_csv(conversion_csv)
        valid = validation.validate_columns(cur, CURRENCY_DATATYPES.keys())
        valid = validation.validate_datatypes(cur, CURRENCY_DATATYPES)
        logger.debug(f"passed validation: {valid}")
        cur["Date"] = pd.to_datetime(cur["Date"])
        cur = cur.set_index("Date")
        cur = cur.rename({"Close": "Convert"}, axis=1)
        cur = cur[["Convert"]]
        self.gbpusd = cur
        return cur

    def calculate_returns(
        self, prices: pd.DataFrame, code: str, convert: bool, cur: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Take in a dataframe of prices of all codes and calculate the returns
        shown by each security and then record those returns into a single
        dataframe.

        The dataframe will be in the format as follows:

        | TIDM | Date | Open | High | Low | Close | Volume | Adjustment |
        |------|------|------|------|-----|-------|--------|------------|

        The function will work by filtering the input frame by each TIDM and
        calculating the adjusted and converted returns and appending them to
        a new dataframe.

        :returns: A dataframe of a single TIDM with returns adjusted by
        adjustment and converted to currency
        """
        ti = prices.loc[prices.TIDM == code]
        if len(ti) < 1:
            raise validation.InvalidDataFrame(f"{code} not in prices df")
        ti = ti.sort_values(by="Date")
        ti = ti.set_index("Date")
        if convert:
            ti = ti.join(cur)
            ti.Close = ti.Close * ti.Convert
        ti["Close"] = ti.Close * ti.Adjustment
        ti["Returns"] = ti.Close.pct_change()
        ti["Returns"] = ti["Returns"].dropna()
        logger.info(f"Calculation for {code} finished")
        return ti

    def analyse(self) -> typing.Dict[str, float]:
        analysis_results = {}

        daily_returns = analysis.calculate_mean_daily_returns(self.clean_returns)
        annual_returns = analysis.calculate_mean_annual_return(self.clean_returns)
        daily_std = analysis.calculate_std_daily(self.clean_returns)
        daily_var = analysis.calculate_variance_daily(self.clean_returns)
        skew = analysis.calculate_skewness(self.clean_returns)
        kurtosis = analysis.calculate_kurtosis(self.clean_returns)
        percentage_returns = analysis.calculate_percentage_returns(self.clean_returns)

        analysis_results["daily_returns"] = daily_returns
        analysis_results["annual_returns"] = annual_returns
        analysis_results["daily_std"] = daily_std
        analysis_results["daily_var"] = daily_var
        analysis_results["skew"] = skew
        analysis_results["kurtosis"] = kurtosis

        self.analysis = analysis_results
        self.percentage_returns = percentage_returns

        logger.info("Analysis loaded")
        return analysis_results

    def plot_correlation_heatmap(self) -> None:
        if len(self.backtest) < 1:
            logger.warn("please run `.build()` before plotting")
        stock_returns = self.backtest.drop("portfolio_returns", axis=1)
        correlation_matrix = stock_returns.corr()
        logger.info("calculating portfolio correlation")
        plot.plot_heatmap(correlation_matrix, "Portfolio Correlation")

    def plot_returns_data(self) -> None:
        if len(self.backtest) < 1:
            logger.warn("please run `.build()` before plotting")
        plot.plot_histogram(self.clean_returns, self.percentage_returns, "Returns data")
